import os
import re
import asyncio


def target_directory(output_path: str | None = None) -> str:
    if output_path:
        if not os.path.isabs(output_path):
            output_path = os.path.join(os.getcwd(), output_path)
    else:
        output_path = os.getcwd()
    os.makedirs(output_path, exist_ok=True)
    return output_path


def safe_filename(s: str, max_length: int = 255) -> str:
    """Sanitize a string making it safe to use as a filename.

    This function was based off the limitations outlined here:
    https://en.wikipedia.org/wiki/Filename.

    :param str s:
        A string to make safe for use as a file name.
    :param int max_length:
        The maximum filename character length.
    :rtype: str
    :returns:
        A sanitized string.
    """
    # Characters in range 0-31 (0x00-0x1F) are not allowed in ntfs filenames.
    ntfs_characters = [chr(i) for i in range(0, 31)]
    characters = [
        r'"',
        r"\#",
        r"\$",
        r"\%",
        r"'",
        r"\*",
        r"\,",
        r"\.",
        r"\/",
        r"\:",
        r'"',
        r"\;",
        r"\<",
        r"\>",
        r"\?",
        r"\\",
        r"\^",
        r"\|",
        r"\~",
        r"\\\\",
    ]
    pattern = "|".join(ntfs_characters + characters)
    regex = re.compile(pattern, re.UNICODE)
    filename = regex.sub("", s)
    return filename[:max_length].rsplit(" ", 0)[0]


def retry_if_none(max_retries: int):
    def decorator(func):
        async def wrapper(*args, **kwargs):
            retries = 0
            while True:
                result = await func(*args, **kwargs)
                if result is not None:
                    return result
                retries += 1
                if retries > max_retries:
                    raise ValueError(f"{func.__name__} failed after {max_retries} retries")
                await asyncio.sleep(1)
        return wrapper
    return decorator



import ast
import json
import re
from urllib.parse import parse_qs, urlparse

from aiotube.exceptions import HTMLParseError, RegexMatchError


def extract_video_id(url: str):
    pattern = r"(?:v=|\/)([0-9A-Za-z_-]{11}).*"
    result = re.search(pattern=pattern, string=url)
    return result.group(1)


def find_object_from_startpoint(html, start_point):
    html = html[start_point:]
    if html[0] not in ["{", "["]:
        raise HTMLParseError(f"Invalid start point. Start of HTML:\n{html[:20]}")

    # First letter MUST be an open brace, so we put that in the stack,
    # and skip the first character.
    stack = [html[0]]
    i = 1

    context_closers = {"{": "}", "[": "]", '"': '"'}

    while i < len(html):
        if len(stack) == 0:
            break
        curr_char = html[i]
        curr_context = stack[-1]

        # If we've reached a context closer, we can remove an element off the stack
        if curr_char == context_closers[curr_context]:
            stack.pop()
            i += 1
            continue

        # Strings require special context handling because they can contain
        #  context openers *and* closers
        if curr_context == '"':
            # If there's a backslash in a string, we skip a character
            if curr_char == "\\":
                i += 2
                continue
        else:
            # Non-string contexts are when we need to look for context openers.
            if curr_char in context_closers.keys():
                stack.append(curr_char)

        i += 1

    full_obj = html[:i]
    return full_obj  # noqa: R504


def parse_for_object_from_startpoint(html, start_point):
    full_obj = find_object_from_startpoint(html, start_point)
    try:
        return json.loads(full_obj)
    except json.decoder.JSONDecodeError:
        try:
            return ast.literal_eval(full_obj)
        except (ValueError, SyntaxError):
            raise HTMLParseError("Could not parse object.")


def parse_for_object(html, preceding_regex):
    regex = re.compile(preceding_regex)
    result = regex.search(html)
    if not result:
        raise HTMLParseError(f"No matches for regex {preceding_regex}")

    start_index = result.end()
    return parse_for_object_from_startpoint(html, start_index)


def initial_player_response(watch_html: str) -> str | dict:
    patterns = [
        r"window\[['\"]ytInitialPlayerResponse['\"]]\s*=\s*",
        r"ytInitialPlayerResponse\s*=\s*",
    ]
    for pattern in patterns:
        try:
            return parse_for_object(watch_html, pattern)
        except HTMLParseError:
            pass

    raise RegexMatchError(
        caller="initial_player_response", pattern="initial_player_response_pattern"
    )


def playability_status(watch_html: str):
    player_response = initial_player_response(watch_html)
    status_dict = player_response.get("playabilityStatus", {})
    if "liveStreamability" in status_dict:
        return "LIVE_STREAM", "Video is a live stream."
    if "status" in status_dict:
        if "reason" in status_dict:
            return status_dict["status"], [status_dict["reason"]]
        if "messages" in status_dict:
            return status_dict["status"], status_dict["messages"]
    return None, [None]


def mime_type_codec(mime_type_codec: str) -> tuple[str, list[str]]:
    pattern = r"(\w+\/\w+)\;\scodecs=\"([a-zA-Z-0-9.,\s]*)\""
    regex = re.compile(pattern)
    results = regex.search(mime_type_codec)
    mime_type, codecs = results.groups()
    return mime_type, [c.strip() for c in codecs.split(",")]


def apply_descrambler(stream_data: dict) -> list | None:
    if "url" in stream_data:
        return None
    formats = []
    if "formats" in stream_data.keys():
        formats.extend(stream_data["formats"])
    if "adaptiveFormats" in stream_data.keys():
        formats.extend(stream_data["adaptiveFormats"])
    for data in formats:
        if "url" not in data:
            if "signatureCipher" in data:
                cipher_url = parse_qs(data["signatureCipher"])
                data["url"] = cipher_url["url"][0]
                data["s"] = cipher_url["s"][0]
        data["is_otf"] = data.get("type") == "FORMAT_STREAM_TYPE_OTF"

    return formats


def extract_playlist_id(url: str) -> str:
    parsed = urlparse(url)
    return parse_qs(parsed.query)["list"][0]


def extract_playlist_info(watch_html: str) -> str | dict:
    patterns = [r"window\[['\"]ytInitialData['\"]]\s*=\s*", r"ytInitialData\s*=\s*"]
    for pattern in patterns:
        try:
            return parse_for_object(watch_html, pattern)
        except HTMLParseError:
            pass

    raise RegexMatchError(
        caller="extract_playlist_info", pattern="extract_playlist_info_pattern"
    )


def extract_video_id_from_playlist(playlist_info: dict):
    section_contents = playlist_info["contents"]["twoColumnBrowseResultsRenderer"][
        "tabs"
    ][0]["tabRenderer"]["content"]["sectionListRenderer"]["contents"]
    important_content = section_contents[0]["itemSectionRenderer"]["contents"][0][
        "playlistVideoListRenderer"
    ]
    videos = important_content["contents"]
    video_ids = list(map(lambda x: x["playlistVideoRenderer"]["videoId"], videos))
    return video_ids



import os
from datetime import datetime
from io import BytesIO
from typing import AsyncGenerator, Callable, List, Optional

from pydantic import BaseModel, HttpUrl

from .client import RequestClient, HttpMethod
from .extractors import mime_type_codec
from .helpers import safe_filename, target_directory
from .itags import get_format_profile


class Stream(BaseModel):
    title: str
    author: str
    url: HttpUrl
    itag: int
    mimeType: str
    bitrate: int
    width: Optional[int]
    height: Optional[int]
    lastModifed: Optional[datetime]
    contentLength: Optional[str]
    quality: str
    fps: Optional[int]
    qualityLabel: Optional[str]
    projectionType: str
    averageBitrate: Optional[int]
    audioQuality: Optional[str]
    approxDurationMs: str
    audioSampleRate: Optional[str]
    audioChannels: Optional[int]
    is_otf: bool

    @property
    def itag_profile(self):
        return get_format_profile(self.itag)

    @property
    def abr(self):
        return self.itag_profile["abr"]

    @property
    def resolution(self):
        return self.itag_profile["resolution"]

    @property
    def mime_type(self):
        mime_types, codecs = mime_type_codec(self.mimeType)
        return mime_types

    @property
    def codecs(self):
        mime_type, codecs = mime_type_codec(self.mimeType)
        return codecs

    @property
    def type(self):
        return self.mimeType.split("/")[0]

    @property
    def subtype(self):
        return self.mime_type.split("/")[1]

    @property
    def is_adaptive(self) -> bool:
        """Whether the stream is DASH.

        :rtype: bool
        """
        return bool(len(self.codecs) % 2)

    @property
    def is_progressive(self) -> bool:
        """Whether the stream is progressive.

        :rtype: bool
        """
        return not self.is_adaptive

    @property
    def includes_audio_track(self) -> bool:
        """Whether the stream only contains audio.

        :rtype: bool
        """
        return self.is_progressive or self.type == "audio"

    @property
    def includes_video_track(self) -> bool:
        """Whether the stream only contains video.

        :rtype: bool
        """
        return self.is_progressive or self.type == "video"

    async def filesize(self) -> int:
        client = RequestClient("ANDROID")
        response = await client.request(method=HttpMethod.HEAD, url=self.url)
        return int(response.get("headers", {}).get("Content-Length"))

    async def _download(self) -> AsyncGenerator[bytes, None]:
        client = RequestClient("ANDROID")
        default_range_size = 9437184
        file_size: int = default_range_size
        downloaded = 0
        while downloaded < file_size:
            stop_pos = min(downloaded + default_range_size, file_size) - 1
            range_header = f"bytes={downloaded}-{stop_pos}"
            request = await client.request(
                method=HttpMethod.GET, url=self.url, headers={"Range": range_header}
            )
            headers = request.get("headers")
            chunk = request.get("response")
            if file_size == default_range_size:
                content_range = headers.get("Content-Range")
                file_size = int(content_range.split("/")[1])
            downloaded += len(chunk)
            yield chunk

    async def download_buffer(self) -> BytesIO:
        io = BytesIO()
        async for chunk in self._download():
            io.write(chunk)
        io.seek(0)
        return io

    async def download_filepath(self, filename: str = None, output_path: str = None):
        file_path = self.get_file_path(
            filename=filename,
            output_path=output_path,
        )
        with open(file_path, "wb") as file:
            async for chunk in self._download():
                file.write(chunk)
        return file_path

    @property
    def default_filename(self) -> str:
        """Generate filename based on the video title.

        :rtype: str
        :returns:
            An os file system compatible filename.
        """
        filename = safe_filename(self.title)
        return f"{filename}.{self.subtype}"

    def get_file_path(
        self,
        filename: Optional[str] = None,
        output_path: Optional[str] = None,
    ) -> str:
        if not filename:
            filename = self.default_filename
        return os.path.join(target_directory(output_path), filename)


class StreamQuery:
    def __init__(self, fmt_streams: List[Stream]) -> None:
        self.fmt_streams = fmt_streams

    def filter(
        self,
        fps=None,
        res=None,
        resolution=None,
        mime_type=None,
        type=None,
        subtype=None,
        file_extension=None,
        abr=None,
        bitrate=None,
        video_codec=None,
        audio_codec=None,
        only_audio=None,
        only_video=None,
        progressive=None,
        adaptive=None,
        is_dash=None,
        custom_filter_functions=None,
    ):
        filters = []
        if res or resolution:
            filters.append(lambda s: s.resolution == (res or resolution))

        if fps:
            filters.append(lambda s: s.fps == fps)

        if mime_type:
            filters.append(lambda s: s.mime_type == mime_type)

        if type:
            filters.append(lambda s: s.type == type)

        if subtype or file_extension:
            filters.append(lambda s: s.subtype == (subtype or file_extension))

        if abr or bitrate:
            filters.append(lambda s: s.abr == (abr or bitrate))

        if video_codec:
            filters.append(lambda s: s.video_codec == video_codec)

        if audio_codec:
            filters.append(lambda s: s.audio_codec == audio_codec)

        if only_audio:
            filters.append(
                lambda s: (s.includes_audio_track and not s.includes_video_track),
            )

        if only_video:
            filters.append(
                lambda s: (s.includes_video_track and not s.includes_audio_track),
            )

        if progressive:
            filters.append(lambda s: s.is_progressive)

        if adaptive:
            filters.append(lambda s: s.is_adaptive)

        if custom_filter_functions:
            filters.extend(custom_filter_functions)

        if is_dash is not None:
            filters.append(lambda s: s.is_dash == is_dash)

        return self._filter(filters)

    def _filter(self, filters: List[Callable]) -> "StreamQuery":
        fmt_streams = self.fmt_streams
        for filter_lambda in filters:
            fmt_streams = filter(filter_lambda, fmt_streams)
        return StreamQuery(list(fmt_streams))

    def order_by(self, attribute_name: str) -> "StreamQuery":
        has_attribute = [
            s for s in self.fmt_streams if getattr(s, attribute_name) is not None
        ]
        if has_attribute and isinstance(getattr(has_attribute[0], attribute_name), str):
            try:
                return StreamQuery(
                    sorted(
                        has_attribute,
                        key=lambda s: int(
                            "".join(filter(str.isdigit, getattr(s, attribute_name)))
                        ),  # type: ignore  # noqa: E501
                    )
                )
            except ValueError:
                pass
        return StreamQuery(
            sorted(has_attribute, key=lambda s: getattr(s, attribute_name))
        )

    def last(self):
        try:
            return self.fmt_streams[-1]
        except IndexError:
            pass

    def get_audio_only(self, subtype: str = "mp4") -> Optional[Stream]:
        return self.filter(only_audio=True, subtype=subtype).order_by("abr").last()

    def get_highest_resolution(self) -> Optional[Stream]:
        return self.filter(progressive=True).order_by("resolution").last()



import json
from enum import Enum
from urllib.parse import urlencode

from aiohttp import ClientSession
from aiohttp.client_exceptions import ContentTypeError

from aiotube.constants import default_clients


class HttpMethod(Enum):
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"
    HEAD = "HEAD"


class RequestClient:
    def __init__(self, client: str):
        self.context = default_clients[client]["context"]
        self.api_key = default_clients[client]["api_key"]
        self.client_data = default_clients[client]

    @property
    def base_params(self) -> dict:
        return {"key": self.api_key, "contentCheckOk": True, "racyCheckOk": True}

    @property
    def base_data(self) -> dict:
        self.context.update(
            {
                "androidSdkVersion": 30,
                "userAgent": self.client_data.get("useragent"),
                "hl": "en",
                "timeZone": "UTC",
                "utcOffsetMinutes": 0
            }
        )
        data = {"context": self.context}
        data.update(
            {
                "params": "8AEB",
                "playbackContext": {
                    "contentPlaybackContext":
                        {"html5Preference": "HTML5_PREF_WANTS"}
                    },
                "contentCheckOk": True,
                "racyCheckOk": True}
        )
        return data

    @property
    def base_headers(self) -> dict:
        client_version = self.context.get("client").get("clientVersion")
        headers = {
            "User-Agent": self.client_data.get("useragent"),
            "accept-language": "en-US,en",
            "content-type": "application/json"
        }
        headers.update(
            {
                "X-YouTube-Client-Name": "3",
                "X-YouTube-Client-Version": client_version,
                "Origin": "https://www.youtube.com"
            }
        )
        return headers

    async def request(
        self,
        method: HttpMethod,
        url: str,
        headers: dict = None,
        params: dict = None,
        data: dict = None,
    ):
        base_headers = self.base_headers
        if params:
            params = urlencode(params)
        if headers:
            base_headers.update(headers)
        if data:
            data = json.dumps(data).encode("utf-8")
        async with ClientSession() as session:
            async with session.request(
                method=method.value,
                url=url,
                headers=headers,
                params=params,
                data=data,
            ) as response:
                headers = response.headers
                try:
                    response_data = await response.json()
                except ContentTypeError:
                    response_data = await response.read()
        return dict(response=response_data, headers=headers)



from aiotube.client import HttpMethod, RequestClient
from aiotube.exceptions import (LiveStreamError, MembersOnly,
                                RecordingUnavailable, VideoPrivate,
                                VideoUnavailable)
from aiotube.extractors import (apply_descrambler, extract_video_id,
                                playability_status)
from aiotube.streams import Stream, StreamQuery
from aiotube.helpers import retry_if_none


class Video:
    def __init__(self, url: str):
        self.video_id = extract_video_id(url)
        self.base_url = "https://www.youtube.com/youtubei/v1"
        self.watch_url = f"https://youtube.com/watch?v={self.video_id}"
        self.client = RequestClient("ANDROID")
        self._html = None
        self._video_info = None
        self._fmt_streams = None

    def __repr__(self):
        return f"<aiotube.video.Video object: videoId={self.video_id}>"

    async def html(self):
        if self._html is None:
            self._html = (
                await self.client.request(method=HttpMethod.GET, url=self.watch_url)
            ).get("response")
        return self._html

    async def video_info(self):
        endpoint = f"{self.base_url}/player"
        query = {"videoId": self.video_id}
        query.update(self.client.base_params)
        data = self.client.base_data
        data.update(
            {"videoId": self.video_id}
        )
        response = await self.client.request(
            method=HttpMethod.POST,
            url=endpoint,
            params=query,
            data=data,
        )
        return response.get("response")

    @retry_if_none(max_retries=5)
    async def streaming_data(self):
        await self.check_availability()
        data = await self.video_info()
        if "streamingData" in data:
            return data["streamingData"]
        return await self.bypass_age_gate()

    async def fmt_streams(self):
        await self.check_availability()
        if self._fmt_streams:
            return self._fmt_streams
        self._fmt_streams = []
        stream_manifest = apply_descrambler(await self.streaming_data())
        video_title = await self.title()
        video_author = await self.author()
        for stream in stream_manifest:
            video = Stream(title=video_title, author=video_author, **stream)
            self._fmt_streams.append(video)
        return self._fmt_streams

    async def streams(self):
        return StreamQuery(await self.fmt_streams())

    async def bypass_age_gate(self):
        client = RequestClient("ANDROID_EMBED")
        endpoint = f"{self.base_url}/player"
        query = {"videoId": self.video_id}
        query.update(self.client.base_params)
        headers = {"Content-Type": "application/json"}
        response = await client.request(
            method=HttpMethod.POST,
            url=endpoint,
            params=query,
            headers=headers,
            data=self.client.base_data,
        )
        return response.get("response")

    async def title(self) -> str:
        """Get the video title."""
        await self.check_availability()
        return (await self.video_info()).get("videoDetails", {}).get("title")

    async def author(self) -> str:
        """Get the video author."""
        return (
            (await self.video_info()).get("videoDetails", {}).get("author", "unknown")
        )

    async def length(self) -> int:
        """Get the video length in seconds."""
        return int(
            (await self.video_info()).get("videoDetails", {}).get("lengthSeconds")
        )

    async def check_availability(self):
        html = (await self.html()).decode()
        status, messages = playability_status(html)
        for reason in messages:
            if status == "UNPLAYABLE":
                if reason == (
                    "Join this channel to get access to members-only content "
                    "like this video, and other exclusive perks."
                ):
                    raise MembersOnly(video_id=self.video_id)
                elif reason == "This live stream recording is not available.":
                    raise RecordingUnavailable(video_id=self.video_id)
                else:
                    raise VideoUnavailable(video_id=self.video_id)
            elif status == "LOGIN_REQUIRED":
                if reason == (
                    "This is a private video. "
                    "Please sign in to verify that you may see it."
                ):
                    raise VideoPrivate(video_id=self.video_id)
            elif status == "ERROR":
                if reason == "Video unavailable":
                    raise VideoUnavailable(video_id=self.video_id)
            elif status == "LIVE_STREAM":
                raise LiveStreamError(video_id=self.video_id)



from re import Pattern


class AiotubeError(BaseException):
    """Base Exception"""


class HTMLParseError(AiotubeError):
    pass


class RegexMatchError(AiotubeError):
    """Regex pattern did not return any matches."""

    def __init__(self, caller: str, pattern: str | Pattern):
        """
        :param str caller:
            Calling function
        :param str pattern:
            Pattern that failed to match
        """
        super().__init__(f"{caller}: could not find match for {pattern}")
        self.caller = caller
        self.pattern = pattern


class VideoUnavailable(AiotubeError):
    """Base video unavailable error."""

    def __init__(self, video_id: str):
        """
        :param str video_id:
            A YouTube video identifier.
        """
        self.video_id = video_id
        super().__init__(self.error_string)

    @property
    def error_string(self):
        return f"{self.video_id} is unavailable"


class MembersOnly(VideoUnavailable):
    def __init__(self, video_id: str):
        """
        :param str video_id:
            A YouTube video identifier.
        """
        self.video_id = video_id
        super().__init__(self.video_id)

    @property
    def error_string(self):
        return f"{self.video_id} is a members-only video"


class RecordingUnavailable(VideoUnavailable):
    def __init__(self, video_id: str):
        """
        :param str video_id:
            A YouTube video identifier.
        """
        self.video_id = video_id
        super().__init__(self.video_id)

    @property
    def error_string(self):
        return f"{self.video_id} does not have a live stream recording available"


class LiveStreamError(VideoUnavailable):
    """Video is a live stream."""

    def __init__(self, video_id: str):
        """
        :param str video_id:
            A YouTube video identifier.
        """
        self.video_id = video_id
        super().__init__(self.video_id)

    @property
    def error_string(self):
        return f"{self.video_id} is streaming live and cannot be loaded"


class VideoPrivate(VideoUnavailable):
    def __init__(self, video_id: str):
        """
        :param str video_id:
            A YouTube video identifier.
        """
        self.video_id = video_id
        super().__init__(self.video_id)

    @property
    def error_string(self):
        return f"{self.video_id} is a private video"



from typing import AsyncGenerator, List

from .client import HttpMethod, RequestClient
from .extractors import (extract_playlist_id, extract_playlist_info,
                         extract_video_id_from_playlist)
from .video import Video


class Playlist:
    def __init__(self, url: str) -> None:
        self.client = RequestClient("ANDROID")
        self._input_url = url
        self._playlist_id = None
        self._html = None
        self._playlist_info = None
        self._sidebar_info = None

    @property
    def playlist_id(self) -> str:
        if self._playlist_id:
            return self._playlist_id
        self._playlist_id = extract_playlist_id(self._input_url)
        return self._playlist_id

    @property
    def playlist_url(self) -> str:
        return f"https://www.youtube.com/playlist?list={self.playlist_id}"

    async def playlist_html(self) -> str:
        if self._html:
            return self._html
        request = await self.client.request(
            method=HttpMethod.GET, url=self.playlist_url
        )
        self._html = request.get("response").decode("utf-8")
        return self._html

    async def playlist_info(self) -> dict:
        if self._playlist_info:
            return self._playlist_info
        info = extract_playlist_info(await self.playlist_html())
        self._playlist_info = info
        return self._playlist_info

    async def sidebar_info(self) -> dict:
        if self._sidebar_info:
            return self._sidebar_info
        info = (await self.playlist_info())["sidebar"]["playlistSidebarRenderer"][
            "items"
        ]
        self._sidebar_info = info
        return self._sidebar_info

    async def video_urls(self) -> List[str]:
        result = []
        for video_id in extract_video_id_from_playlist(await self.playlist_info()):
            result.append(self.create_url(video_id))
        return result

    async def video_generator(self) -> AsyncGenerator[Video, None]:
        for video in await self.video_urls():
            yield Video(video)

    def create_url(self, video_id: str) -> str:
        return f"https://www.youtube.com/watch?v={video_id}"



import setuptools


def get_description():
    with open('README.rst', 'r', encoding='utf-8') as f:
        return f.read()


setuptools.setup(
    name="aiotubes",
    version="3.4",
    license='MIT',
    author="sheldy",
    description="Asynchronous Youtube API",
    url="https://github.com/sheldygg/aiotube",
    packages=setuptools.find_packages(),
    long_description_content_type='text/markdown',
    long_description=get_description(),
    classifiers=[
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11"
    ],
    install_requires=[
        "aiohttp",
        "pydantic"
    ],)
